-- prepared by Steve Stedman
-- http://SteveStedman.com
-- twitter:  @SqlEmt
-- Logical Functions

RAISERROR ('Dont run the whole script, just run sections at a time', 20, 1)  WITH LOG;


USE Master;
GO

IF EXISTS(SELECT name 
            FROM sys.databases 
		   WHERE name = 'QueryTraining')
BEGIN
	ALTER DATABASE [QueryTraining] 
	  SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
	DROP DATABASE [QueryTraining];
END
CREATE DATABASE [QueryTraining];
GO

USE [QueryTraining];
GO

------------------------ EXAMPLES SETUP ------------------------
-- Table to be used for training
CREATE TABLE REVENUE
(
[DepartmentID] int,
[Revenue] int,
[Year] int
);
 
insert into REVENUE
values (1,10030,1998),(2,20000,1998),(3,40000,1998),
(1,20000,1999),(2,60000,1999),(3,50000,1999),
(1,40000,2000),(2,40000,2000),(3,60000,2000),
(1,30000,2001),(2,30000,2001),(3,70000,2001),
(1,90000,2002),(2,20000,2002),(3,80000,2002),
(1,10300,2003),(2,1000,2003), (3,90000,2003),
(1,10000,2004),(2,10000,2004),(3,10000,2004),
(1,20000,2005),(2,20000,2005),(3,20000,2005),
(1,40000,2006),(2,30000,2006),(3,30000,2006),
(1,70000,2007),(2,40000,2007),(3,40000,2007),
(1,50000,2008),(2,50000,2008),(3,50000,2008),
(1,20000,2009),(2,60000,2009),(3,60000,2009),
(1,30000,2010),(2,70000,2010),(3,70000,2010),
(1,80000,2011),(2,80000,2011),(3,80000,2011),
(1,10000,2012),(2,90000,2012),(3,90000,2012);
 






------------------------------------------------
-- CASE
USE [QueryTraining];
GO



declare @corners as int = 6;
-- the old way using case.
SELECT CASE @corners
		WHEN 1 THEN 'point'
		WHEN 2 THEN 'line'
		WHEN 3 THEN 'triangle'
		WHEN 4 THEN 'square'
		WHEN 5 THEN 'pentagon'
		WHEN 6 THEN 'hexagon'
		WHEN 7 THEN 'heptagon'
		WHEN 8 THEN 'octagon'
		ELSE NULL
	   END;


-- http://stevestedman.com/?p=1578
-- assume we want to display an indicator to see if we are above
-- or below average. First we start with the average over departmentID
SELECT Year,
       DepartmentID,
       Revenue,
       AVG(Revenue) OVER (PARTITION BY DepartmentID) AS AverageRevenue
  FROM REVENUE
 ORDER BY DepartmentID, year;

-- using the CASE statement we would get the following
-- http://stevestedman.com/?p=1578
SELECT Year, DepartmentID, Revenue, AverageRevenue,
       CASE WHEN Revenue > AverageRevenue
            THEN 'Better Than Average'
            ELSE 'Not Better'
        END AS Ranking
  FROM (SELECT Year, DepartmentID, Revenue,
               avg(Revenue) OVER (PARTITION by DepartmentID) as AverageRevenue
          FROM REVENUE 
	   ) as t
 ORDER BY DepartmentID, year;




------------------------------------------------
-- IIF (2012)
USE [QueryTraining];
GO

-- now the same functionality using IIF and simplifying the code
-- http://stevestedman.com/?p=1578
SELECT  Year,
        DepartmentID,
        Revenue,
        AverageRevenue,
		IIF(Revenue > AverageRevenue, 'Better Than Average', 'Not Better' ) AS Ranking
FROM    (SELECT Year,
                DepartmentID,
                Revenue,
                AVG(Revenue) OVER (PARTITION BY DepartmentID) AS AverageRevenue
         FROM   REVENUE
        ) AS t
ORDER BY DepartmentID, year;
 
----------------------------------
SELECT IIF(3 = 2, 'true', NULL); -- ERROR ?
 
----------------------------------
 
SELECT IIF(1 = 7, NULL, NULL); -- ERROR !
--Do you think that IIF actually uses the CASE statement?


------------------------------------------------
-- CHOOSE (2012)
-- returns the item at a specific index
-- http://stevestedman.com/?p=1691

declare @corners as int = 6;
-- the old way using case.
SELECT CASE @corners
		WHEN 1 THEN 'point'
		WHEN 2 THEN 'line'
		WHEN 3 THEN 'triangle'
		WHEN 4 THEN 'square'
		WHEN 5 THEN 'pentagon'
		WHEN 6 THEN 'hexagon'
		WHEN 7 THEN 'heptagon'
		WHEN 8 THEN 'octagon'
		ELSE NULL
	   END;

-- now with choose
declare @corners as int = 6;
SELECT choose(@corners, 'point', 'line', 'triangle', 'square', 'pentagon', 'hexagon', 'heptagon', 'octagon')




-- CHOOSE day of week example
DECLARE @day as int=4
SELECT CHOOSE(@day,'Sunday','Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday');


-- Pick 5 people at random from a list of 10
GO
DECLARE @myRandonNum INTEGER;
SET @myRandonNum = CAST(RAND() * 10 AS INTEGER) + 1;
SELECT @myRandonNum as Num, CHOOSE(@myRandonNum, 'Steve', 'Aaron', 'Pete', 'Cody', 'Gabe', 'Bill', 'Aaron', 'Jordan', 'Darren', 'Mark', 'Aaron') as Name;
GO 5



------------------------------------------------
-- COALESCE

USE [QueryTraining];

CREATE TABLE [dbo].[Departments](
	[ID] [INTEGER] IDENTITY(1,1) PRIMARY KEY NOT NULL,
	[Department] [VARCHAR](200) NOT NULL,
	[Details] [VARCHAR](4000) NULL,
	[Parent] [INT] NULL,
	[Archived] [BIT] NOT NULL DEFAULT(0)
);

INSERT INTO [dbo].[Departments]
           ([Department])
     VALUES
           ('Camping'),
           ('Cycle'),
           ('Snowsports'),
           ('Fitness');


DECLARE @campingDepartmentId as INT;
SELECT @campingDepartmentId = ID
  FROM [dbo].[Departments]
  WHERE [Department] = 'Camping';

INSERT INTO [dbo].[Departments]
           ([Department], [Parent])
     VALUES
           ('Tents', @campingDepartmentId),
           ('Backpacks', @campingDepartmentId),
           ('Sleeping Bags', @campingDepartmentId),
           ('Cooking', @campingDepartmentId);


-- examine the table
SELECT [ID],
	   [Department],
       [Details],
       [Parent],
       [Archived]
  FROM [dbo].[Departments];


-- clean up some NULLs with COALESCE
SELECT [ID],
	   [Department],
       [Details],
       COALESCE([Parent], 0) as Parent,
       [Archived]
  FROM [dbo].[Departments];




-- COALESCE compared to ISNULL
-- turn on actual execution plan
set statistics io on;
SELECT [ID],
	   [Department],
       [Details],
	   [Parent],
       COALESCE((SELECT MAX(Parent) FROM [dbo].[Departments]), 0) as MaxParent,
       [Archived]
  FROM [dbo].[Departments];
SELECT [ID],
	   [Department],
       [Details],
	   [Parent],
       ISNULL((SELECT MAX(Parent) FROM [dbo].[Departments]), 0) as MaxParent,
       [Archived]
  FROM [dbo].[Departments];







