-- prepared by Aaron Buma
-- http://AaronBuma.com
-- twitter:  @AaronDBuma
-- User Defined Functions

RAISERROR ('Dont run the whole script, just run sections at a time', 20, 1)  WITH LOG;

------------------------ EXAMPLES SETUP ------------------------

USE Master;
GO

IF EXISTS(SELECT name 
            FROM sys.databases 
		   WHERE name = 'QueryTraining')
BEGIN
	ALTER DATABASE [QueryTraining] 
	  SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
	DROP DATABASE [QueryTraining];
END
CREATE DATABASE [QueryTraining];
GO

USE [QueryTraining];


CREATE TABLE [dbo].[Inventory](
	[InventoryID] [INTEGER] IDENTITY(-2147483647,1) PRIMARY KEY,
	[CategoryID] [INT], 
	[ItemSKU] [VARCHAR](200),
	[Description] [VARCHAR](4000),
	[QuantityOnHand] [INT],
	[UnitCost] DECIMAL(10,2),
	[UnitRetail] DECIMAL(10,2),
	[Archived] BIT
); 


CREATE TABLE [dbo].[Categories](
	[CategoryID] [INTEGER],
	[CategoryTitle] [VARCHAR](200),
	[Description] [VARCHAR](200),
	[Archived] BIT
); 

INSERT INTO dbo.Categories ( CategoryID, CategoryTitle, Description, Archived )
VALUES  
(1, 'SLEEPINGBAG','Body Socks', 0),
(2, 'TENT','Suspended Tarps', 0),
(3, 'AXE','Finger Removers', 0),
(4, 'FOOD','Animal Bait', 0);

INSERT INTO dbo.Inventory (CategoryID, ItemSKU, Description, QuantityOnHand, UnitCost, UnitRetail, Archived)
VALUES
(1, 'SB001-LG-RED', 'Sleeping Bag - Large - RED', 5, 5.75, 19.99, 0),
(1, 'SB001-LG-GRN', 'Sleeping Bag - Large - GREEN', 13, 5.75, 14.99, 1),
(1, 'SB001-LG-BLK', 'Sleeping Bag - Large - BLACK', 1, 5.75, 19.99, 0),
(2, 'TT001-10-CMO', 'Tent - 10 Man - CAMO', 5, 10.00, 99.95, 0),
(3, 'AX013-HT-BLK', 'Axe - Hatchet - Black', 35, 4.99, 24.95, 0),
(4, 'MR005-1S-SPG', 'MRE - 1 Serve - Spaghetti', 5, 1.25, 7.99, 0),
(4, 'MR006-1A-MYS', 'MRE - 1 Serve - Mystery Meat', 1, 1.25, 9.95, 0);
USE [QueryTraining];
  
------------------------------------------------
-- Simple UDF (User Defined Function)

--How many of a given category do we have on hand?
DECLARE @category INT = 1;

SELECT c.CategoryTitle
  FROM dbo.Categories AS c
 WHERE c.CategoryID = @category;

--How can we encapsulate that logic for re-use in multiple places?
GO
CREATE FUNCTION dbo.NameForCategory (@category INT)
RETURNS VARCHAR(200)
AS
BEGIN
	DECLARE @categoryName VARCHAR(200);

	SELECT @categoryName = c.CategoryTitle
	  FROM dbo.Categories AS c
	 WHERE c.CategoryID = @category;

	RETURN @categoryName;
END

/*   Ctrl+Shift+R - Refresh IntelliSense Cache */
SELECT dbo.NameForCategory(1);

-------------------------------------------------
--Alter
ALTER FUNCTION dbo.NameForCategory (@category INT)
RETURNS VARCHAR(200)
AS
BEGIN
	DECLARE @categoryName VARCHAR(200);

	SELECT @categoryName = c.CategoryTitle
	  FROM dbo.Categories AS c
	 WHERE c.CategoryID = @category;

	RETURN @categoryName;
END
-------------------------------------------------
--DROP
DROP FUNCTION dbo.NameForCategory


--RECREATE
CREATE FUNCTION dbo.NameForCategory (@category INT)
RETURNS VARCHAR(200)
AS
BEGIN
	DECLARE @categoryName VARCHAR(200);

	SELECT @categoryName = c.CategoryTitle
	  FROM dbo.Categories AS c
	 WHERE c.CategoryID = @category;

	RETURN @categoryName;
END 

-------------------------------------------------
--Performance 

--Compare Execution plans:
SET STATISTICS IO ON;
SET STATISTICS TIME ON;

--Using a join to get the data 
SELECT c.CategoryTitle
	FROM dbo.Inventory AS i
	INNER JOIN dbo.Categories AS c ON c.CategoryID = i.CategoryID;

--Using a function; should be fast right? Execution plan says so!
SELECT dbo.NameForCategory(i.CategoryID)
	FROM dbo.Inventory AS i;
--But look at the 'compute scalar' property
--DANGER! The Scalar function is a 'black box' and not included in the compiled execution plan
--Scalar Functions and Multi-statement table value functionst are black boxes.



--Slides
-------------------------------------------------
--Table Valued Functions 
--Single-statement TVF
CREATE FUNCTION dbo.CategoryTitleandID ()
RETURNS TABLE --<-- 
AS 
RETURN
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c;
	  
--Return all rows	  
SELECT * FROM dbo.CategoryTitleandID ();

--Filter rows, notice it shows in the execution plan?
SELECT *
  FROM dbo.CategoryTitleandID()
 WHERE CategoryTitle = 'SLEEPINGBAG';

--You can join on it
SELECT *
  FROM dbo.CategoryTitleandID() c
 INNER JOIN dbo.Inventory AS i ON i.CategoryID = c.CategoryID

-------------------------------------------------
--Multi-statement TVF
--SYNTAX is different

CREATE FUNCTION dbo.CategoryTitleandIDMULTI ()
RETURNS @CategoryTitleForID TABLE			/*   NOTICE IT's NOT "RETURNS TABLE"     */
(
	CategoryTitle VARCHAR(200) NOT NULL,
	CategoryID INT NOT NULL
)
AS 
BEGIN										/*   NOTICE IT's NOT "RETURN"     */
	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c ;
/*
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
	DELETE @CategoryTitleForID;

	INSERT INTO @CategoryTitleForID (CategoryTitle, CategoryID)
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c 
*/
	 RETURN
END


SELECT *
  FROM dbo.CategoryTitleandID() c
 INNER JOIN dbo.Inventory AS i ON i.CategoryID = c.CategoryID;
SELECT *
  FROM dbo.CategoryTitleandIDMULTI() c
 INNER JOIN dbo.Inventory AS i ON i.CategoryID = c.CategoryID;
  

--Slides
------------------------------------------------
--Views
--Adding Deterministic functions
CREATE VIEW vCategoryTitleShorthand
WITH SCHEMABINDING
AS
	SELECT LEFT(c.CategoryTitle,6) AS ShortTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c;
GO

SELECT *
  FROM vCategoryTitleShorthand

--The ShortTitle field will be used a lot for queries
CREATE UNIQUE CLUSTERED INDEX
vCategoryTitleShorthand_ShortTitle
ON dbo.vCategoryTitleShorthand (ShortTitle);

------------------------------------------------
--Adding Non-Deterministic functions
CREATE VIEW vInventoryWithCategoryTitle
WITH SCHEMABINDING
AS
	SELECT c.CategoryTitle,
		   i.CategoryID,
		   i.InventoryID,
		   i.ItemSKU,
		   i.Description,
		   i.QuantityOnHand,
		   i.UnitCost,
		   i.UnitRetail,
		   i.Archived
	  FROM dbo.CategoryTitleandID() c
	 INNER JOIN dbo.Inventory AS i ON i.CategoryID = c.CategoryID
GO

/*
Msg 4513, Level 16, State 2, Procedure vInventoryWithCategoryTitle, Line 268
Cannot schema bind view 'vInventoryWithCategoryTitle'. 'dbo.CategoryTitleandID' is not schema bound.
*/

--The function needs to be schemabound
ALTER FUNCTION dbo.CategoryTitleandID ()
RETURNS TABLE 
WITH SCHEMABINDING--<-- 
AS 
RETURN
	SELECT c.CategoryTitle,
		   c.CategoryID
	  FROM dbo.Categories AS c;

--Now you can create the view
CREATE VIEW vInventoryWithCategoryTitle
WITH SCHEMABINDING
AS
	SELECT c.CategoryTitle,
		   i.CategoryID,
		   i.InventoryID,
		   i.ItemSKU,
		   i.Description,
		   i.QuantityOnHand,
		   i.UnitCost,
		   i.UnitRetail,
		   i.Archived
	  FROM dbo.CategoryTitleandID() c
	 INNER JOIN dbo.Inventory AS i ON i.CategoryID = c.CategoryID
GO

SELECT *
  FROM vInventoryWithCategoryTitle;